﻿using System.Collections.Generic;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;

namespace HeuristicLab.Problems.ProgramSynthesis {
  /// <summary>
  /// Counts the occurrences of the top lit-type item in the top type vector.
  /// </summary>
  /// <typeparam name="T"></typeparam>
  [StorableClass]
  public abstract class VectorOccurrenceOfExpression<T> : StatelessExpression {
    protected VectorOccurrenceOfExpression() { }
    [StorableConstructor]
    protected VectorOccurrenceOfExpression(bool deserializing) : base(deserializing) { }

    protected bool IsNoop(IPushStack<IReadOnlyList<T>> vectorStack, IPushStack<T> literalStack) {
      return vectorStack.IsEmpty || literalStack.IsEmpty;
    }

    protected void Eval(
      IInternalPushInterpreter interpreter,
      IPushStack<IReadOnlyList<T>> vectorStack,
      IPushStack<T> literalStack,
      bool canOverride = false) {
      var vector = vectorStack.Pop();
      var literal = canOverride ? literalStack.Top : literalStack.Pop();

      var occurrence = 0;
      for (var i = 0; i < vector.Count; i++) {
        if (vector[i].Equals(literal)) occurrence++;
      }

      if (canOverride) interpreter.IntegerStack.Top = occurrence;
      else interpreter.IntegerStack.Push(occurrence);
    }
  }

  [StorableClass]
  [PushExpression(
    StackTypes.IntegerVector,
    "INTEGER[].OCCURRENCEOF",
    "Pushes the amount of occurrences of the top INTEGER in the top INTEGER[] onto the INTEGER stack.",
    StackTypes.Integer)]
  public class IntegerVectorOccurrenceOfExpression : VectorOccurrenceOfExpression<long> {
    public IntegerVectorOccurrenceOfExpression() { }
    [StorableConstructor]
    protected IntegerVectorOccurrenceOfExpression(bool deserializing) : base(deserializing) { }

    public override bool IsNoop(IInternalPushInterpreter interpreter) {
      return IsNoop(interpreter.IntegerVectorStack, interpreter.IntegerStack);
    }

    public override void Eval(IInternalPushInterpreter interpreter) {
      Eval(interpreter, interpreter.IntegerVectorStack, interpreter.IntegerStack, true);
    }
  }

  [StorableClass]
  [PushExpression(
    StackTypes.FloatVector,
    "FLOAT[].OCCURRENCEOF",
    "Pushes the amount of occurrences of the top FLOAT in the top FLOAT[] onto the INTEGER stack.",
    StackTypes.Float | StackTypes.Integer)]
  public class FloatVectorOccurrenceOfExpression : VectorOccurrenceOfExpression<double> {
    public FloatVectorOccurrenceOfExpression() { }
    [StorableConstructor]
    protected FloatVectorOccurrenceOfExpression(bool deserializing) : base(deserializing) { }

    public override bool IsNoop(IInternalPushInterpreter interpreter) {
      return IsNoop(interpreter.FloatVectorStack, interpreter.FloatStack);
    }

    public override void Eval(IInternalPushInterpreter interpreter) {
      Eval(interpreter, interpreter.FloatVectorStack, interpreter.FloatStack);
    }
  }

  [StorableClass]
  [PushExpression(
    StackTypes.BooleanVector,
    "BOOLEAN[].OCCURRENCEOF",
    "Pushes the amount of occurrences of the top BOOLEAN in the top BOOLEAN[] onto the INTEGER stack.",
    StackTypes.Boolean | StackTypes.Integer)]
  public class BooleanVectorOccurrenceOfExpression : VectorOccurrenceOfExpression<bool> {
    public BooleanVectorOccurrenceOfExpression() { }
    [StorableConstructor]
    protected BooleanVectorOccurrenceOfExpression(bool deserializing) : base(deserializing) { }

    public override bool IsNoop(IInternalPushInterpreter interpreter) {
      return IsNoop(interpreter.BooleanVectorStack, interpreter.BooleanStack);
    }

    public override void Eval(IInternalPushInterpreter interpreter) {
      Eval(interpreter, interpreter.BooleanVectorStack, interpreter.BooleanStack);
    }
  }

  [StorableClass]
  [PushExpression(
    StackTypes.StringVector,
    "STRING[].OCCURRENCEOF",
    "Pushes the amount of occurrences of the top STRING in the top STRING[] onto the INTEGER stack.",
    StackTypes.String | StackTypes.Integer)]
  public class StringVectorOccurrenceOfExpression : VectorOccurrenceOfExpression<string> {
    public StringVectorOccurrenceOfExpression() { }
    [StorableConstructor]
    protected StringVectorOccurrenceOfExpression(bool deserializing) : base(deserializing) { }

    public override bool IsNoop(IInternalPushInterpreter interpreter) {
      return IsNoop(interpreter.StringVectorStack, interpreter.StringStack);
    }

    public override void Eval(IInternalPushInterpreter interpreter) {
      Eval(interpreter, interpreter.StringVectorStack, interpreter.StringStack);
    }
  }
}
